import math, sys
import django
from django.http import QueryDict

PYTHON_2 = sys.version_info < (3,)
DJANGO_GTE_18 = django.VERSION >= (1, 8)


class Page:
    def __init__(self, current_page, page_number):
        self.is_current = current_page == page_number
        self.number = page_number


class Paginator:
    def __init__(self, first_page, prev_page, page_range, next_page, last_page, current_page):
        pages = list()
        for page_number in page_range:
            pages.append(Page(current_page, page_number))

        self.first = Page(current_page, first_page) if first_page is not None else None
        self.prev = Page(current_page, prev_page) if prev_page is not None else None
        self.pages = pages
        self.next = Page(current_page, next_page) if next_page is not None else None
        self.last = Page(current_page, last_page) if last_page is not None else None


def make_paginator(page_obj, num_links):
    number = page_obj.number
    page_count = len(page_obj.paginator.page_range)

    prev_page = page_obj.previous_page_number() if page_obj.has_previous() else None
    next_page = page_obj.next_page_number() if page_obj.has_next() else None

    # Try to play nice with Python 2
    # Force float division before math.ceil
    middle_point = math.ceil(num_links / 2.0)
    if PYTHON_2:
        # Convert float to int
        middle_point = int(middle_point)

    first_page = page_obj.paginator.page_range[0] if page_count > num_links and number > middle_point else None

    last_page = page_obj.paginator.page_range[-1]

    if num_links % 2 == 1:
        not_last_section = number <= (last_page - middle_point)
    else:
        not_last_section = number < (last_page - middle_point)

    last_page = last_page if page_count > num_links and not_last_section else None

    if first_page is None:
        page_range = page_obj.paginator.page_range[:num_links]
    elif last_page is None:
        page_range = page_obj.paginator.page_range[-num_links:]
    else:
        start = number - (middle_point - 1) - 1  # zero indexed
        end = number + middle_point - 1  # zero indexed

        # fix for even number of links
        if num_links % 2 == 0:
            end += 1

        page_range = page_obj.paginator.page_range[start:end]

    return Paginator(first_page, prev_page, page_range, next_page, last_page, number)


def process_querystring(request, page_kwarg):
    if page_kwarg and (len(request.GET) > 0):
        qs = request.GET.copy()

        if DJANGO_GTE_18 or not isinstance(qs, QueryDict):
            new_qs = QueryDict('', mutable=True)
            new_qs.update(qs)
            qs = new_qs

        if page_kwarg in qs:
            qs.pop(page_kwarg)

        return qs.urlencode()

    return ''



from django.core.paginator import Page

MISSING_FIRST_ARG = 'Missing first argument (instance of "{}")'.format(Page.__name__)
MISSING_SECOND_ARG = 'Missing second argument (number of links)'
MISSING_THIRD_ARG = 'Missing third argument (template variable name)'

WRONG_FIRST_ARG = 'First parameter must be an instance of "{}"'.format(Page.__name__)
WRONG_SECOND_ARG = 'Second parameter must be an instance of "{}"'.format(int.__name__)

WRONG_ARGS = '"{{}}" requires an instance of "{}",' \
             ' the number of links to show' \
             ' and, optionally, the "page_kwarg" name'.format(Page.__name__)


import os
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-smart-pagination',
    version='1.0.1',
    packages=find_packages(exclude=['tests']),
    description='Generate pagination links in Django Templates',
    long_description=long_description,
    url='https://github.com/livioribeiro/django-smart-pagination',
    author='Livio Ribeiro',
    author_email='livioribeiro@outlook.com',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],

    keywords='django pagination page links',
    install_requires=['Django'],
    extras_require={
        'test': ['pytest', 'pytest-cov', 'pytest-django', 'Jinja2'],
    }
)



from django.core.paginator import Page
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.exceptions import TemplateSyntaxError, TemplateError

from .. import pagination, error_messages as errors


class PaginationExtension(Extension):
    tags = {'paginate'}

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        request = nodes.Name('request', 'load')

        args = [request]

        try:
            page_obj = parser.parse_expression()
            args.append(page_obj)
        except TemplateSyntaxError:
            raise TemplateSyntaxError(errors.MISSING_FIRST_ARG, lineno)

        try:
            num_links = parser.parse_expression()
            args.append(num_links)
        except TemplateSyntaxError:
            raise TemplateSyntaxError(errors.MISSING_SECOND_ARG, lineno)

        try:
            var_name = parser.parse_expression().name
        except TemplateSyntaxError:
            raise TemplateSyntaxError(errors.MISSING_THIRD_ARG, lineno)

        try:
            page_kwarg = parser.parse_expression()
            args.append(page_kwarg)
        except TemplateSyntaxError:
            pass

        call_node = self.call_method('_make_paginator', args)
        body = parser.parse_statements(['name:endpaginate'], drop_needle=True)
        body.insert(0, nodes.Assign(nodes.Name(var_name, 'store'), call_node))
        return body

    @staticmethod
    def _make_paginator(request, page_obj, num_links, page_kwarg=None):
        if not isinstance(page_obj, Page):
            raise TemplateError(errors.WRONG_FIRST_ARG)

        if not isinstance(num_links, int):
            raise TemplateError(errors.WRONG_SECOND_ARG)

        paginator = pagination.make_paginator(page_obj, num_links)
        query = pagination.process_querystring(request, page_kwarg) if request else None

        if query:
            paginator.query = query

        return paginator



from django.core.paginator import Page
from django import template
from django.template import TemplateSyntaxError

from .. import pagination, error_messages as errors

register = template.Library()


def paginate(parser, token):
    contents = token.split_contents()

    if len(contents) == 1:
        raise TemplateSyntaxError(errors.MISSING_FIRST_ARG)
    elif len(contents) == 2:
        raise TemplateSyntaxError(errors.MISSING_SECOND_ARG)
    elif len(contents) == 3:
        raise TemplateSyntaxError(errors.MISSING_THIRD_ARG)
    elif len(contents) == 4:
        tag_name, page_obj, num_links, var_name = contents
        page_kwarg = None
    elif len(contents) == 5:
        tag_name, page_obj, num_links, var_name, page_kwarg = contents
    else:
        raise TemplateSyntaxError(errors.WRONG_ARGS.format(contents[0]))

    nodelist = parser.parse(('endpaginate',))
    parser.delete_first_token()

    var_name = var_name.strip('"\'')
    page_kwarg = page_kwarg.strip('"\'') if page_kwarg is not None else None

    return PaginationNode(nodelist, page_obj, num_links, var_name, page_kwarg)


register.tag(paginate)


class PaginationNode(template.Node):
    def __init__(self, nodelist, page_obj, num_links, var_name, page_kwarg):
        self.nodelist = nodelist
        self.page_obj = page_obj
        self.num_links = num_links
        self.var_name = var_name
        self.page_kwarg = page_kwarg

    def render(self, context):
        request = context.get('request', None)

        page_obj = template.Variable(self.page_obj).resolve(context)

        if not isinstance(page_obj, Page):
            raise template.TemplateSyntaxError(errors.WRONG_FIRST_ARG)

        try:
            num_links = int(self.num_links)
        except ValueError:
            if self.num_links.startswith("'") or self.num_links.startswith('"'):
                raise template.TemplateSyntaxError(errors.WRONG_SECOND_ARG)

            num_links = template.Variable(self.num_links).resolve(context)

            if not isinstance(num_links, int):
                raise template.TemplateSyntaxError(errors.WRONG_SECOND_ARG)

        query = pagination.process_querystring(request, self.page_kwarg) if request else None
        paginator = pagination.make_paginator(page_obj, num_links)

        if query:
            paginator.query = query

        context.update({self.var_name: paginator})

        return self.nodelist.render(context)
